@@ -1,7 +1,7 @@
 """Manages sets of data on disk.
 
-Tomate provides ways to manipulate data under the form of a
-multi-dimensional array.
+Tomate provides ways to manipulate data under the form of
+multi-dimensional arrays.
 It manages multiples variables, as well as the coordinates along
 which the data varies.
 It also provides multiple convenience functions to retrieve
@@ -24,22 +24,23 @@ import sys
 import tomate.log
 
 from .coordinates.coord import Coord
+from .coordinates.coord_str import CoordStr
 from .coordinates.time import Time
 from .coordinates.latlon import Lat, Lon
 from .coordinates.variables import Variables
 
 from .keys.keyring import Keyring
 
-from .variables_info import VariablesInfo
 from .data_base import DataBase
 from .constructor import Constructor
 
 
-__version__ = "1.1.0"
+__version__ = "2.0.0"
 
 
 __all__ = [
     'Coord',
+    'CoordStr',
     'Time',
     'Lat',
     'Lon',
@@ -47,7 +48,6 @@ __all__ = [
 
     'Keyring',
 
-    'VariablesInfo',
     'DataBase',
     'Constructor'
 ]
