@@ -40,6 +40,16 @@ class PlotObjectAvgABC(PlotObjectABC):
 
     def _get_data(self):
         if 'DataCompute' not in self.db.bases:
-            raise TypeError("DataComptue necessary for averaging.")
-        data = self.db.mean(self.avg_dims, **self.keyring.kw)
+            raise TypeError("DataCompute necessary for averaging.")
+        keyring = self.keyring.copy()
+        var = self.scope.var.get_str_name(keyring.pop('var').value)
+        data = self.db.mean(var, self.avg_dims, **keyring.kw)
+
+        # Nothing to reorder
+        if self.DIM == 1:
+            return data
+
+        axes = [d for d in self.db[var].dims if d not in self.avg_dims]
+        target = self.axes[:2][::-1]
+        data = self.db[var].acs.reorder(axes, data, target)
         return data
