@@ -7,23 +7,24 @@
 
 
 import logging
-from typing import (Any, Callable, Dict, Iterator, List, Optional,
-                    Tuple, Type, Union, TYPE_CHECKING)
+from typing import (Any, Callable, Dict, Iterable, Iterator, List, Optional,
+                    Type, TYPE_CHECKING)
 
 import os
 import re
 
 import numpy as np
 
-import tomate.filegroup.coord_scan as dlcs
 
 from tomate.coordinates.coord import Coord
 from tomate.custom_types import File
-from tomate.filegroup.coord_scan import CoordScan
+from tomate.filegroup.coord_scan import CoordScan, get_coordscan
+from tomate.filegroup.matcher import Matcher
+from tomate.filegroup.scanner import Scanner
+from tomate.filegroup.spec import CoordScanSpec
 from tomate.keys.key import Key, KeyValue
 from tomate.variables_info import VariablesInfo
 if TYPE_CHECKING:
-    from tomate.filegroup.filegroup_load import FilegroupLoad
     from tomate.data_base import DataBase
 
 log = logging.getLogger(__name__)
@@ -37,9 +38,9 @@ class FilegroupScan():
 
     :param root: Root data directory containing all files.
     :param db: Parent database.
-    :param coords_fg: Parent coordinates objects,
-        a bool indicating if the coordinate is shared accross files,
-        and their name inside files.
+    :param coords_fg: Coordinates present in those files, see
+        :class:`CoordScanSpec <tomate.filegroup.spec.CoordScanSpec>` for
+        details.
     :param vi: Global VariablesInfo instance.
     :param name: [opt] Name of the filegroup.
 
@@ -48,36 +49,33 @@ class FilegroupScan():
     :attr vi: VariablesInfo: Global VariablesInfo instance.
     :attr name: str: Name of the filegroup.
 
-    :attr cs: Dict[str, CoordScan or subclass]: Dictionnary of scanning
-        coordinates, each dynamically inheriting from its parent Coord.
-
     :attr pregex: str: Pre-regex.
     :attr regex: str: Regex.
+    :attr file_override: str: If filegroup consist of a single file, this is
+        used (instead of a regular expression).
 
+    :attr found_file: bool: If any file matching the regex has been found.
+    :attr n_matcher: int: Number of matchers in the pre-regex.
     :attr segments: List[str]: Fragments of filename used for reconstruction,
         elements with pair indices are replaced with matches.
 
-    :attr scan_attr: Dict[str, [Callable, scanned:bool, kwargs: Dict]]:
-        Functions to call to scan variables specific
-        attributes or general attributes.
-        Key is 'gen' for general attributes, 'var' for variable specific.
-        Value is a tuple of the function to call, a boolean if the attributes
-        have been scanned, and kwargs to pass to the function.
-
+    :attr cs: Dict[str, CoordScan]: Dictionnary of scanning coordinates, each
+        dynamically inheriting from its parent Coord.
     :attr selection: Dict[str, Union[KeyLike, KeyLikeValue]]:
-        Keys for selecting parts of the CoordScan, by index or value.
-        Dict key is dimension name.
-
-    :attr post_loading_funcs: List[Tuple[Callable, KeyVar, bool, Dict]]:
+        Keys for selecting parts of each CoordScan, by index or value.
+    :attr post_loading_funcs: List[Tuple[Callable, Key, bool, Dict]]:
         Functions applied after loading data.
         Each element is a tuple of the function, the variable that triggers
         the call, a boolean True if all said variables must present to trigger,
         False if any variable must be loaded, and kwargs to pass.
     """
 
+    MAX_DEPTH_SCAN = 3
+    """Limit descending into lower directories when finding files."""
+
     def __init__(self, root: str,
                  db: 'DataBase',
-                 coords_fg: List[Tuple[Coord, bool, str]],
+                 coords_fg: List[CoordScanSpec],
                  vi: VariablesInfo,
                  name: str = ''):
         self.root = root
@@ -85,14 +83,15 @@ class FilegroupScan():
         self.vi = vi
         self.name = name
 
+        self.regex = ""
+        self.pregex = ""
+        self.file_override = ''
+
         self.found_file = False
         self.n_matcher = 0
         self.segments = []
 
-        self.regex = ""
-        self.pregex = ""
-
-        self.scan_attr = {}
+        self.scanners = []
 
         self.cs = {}
         self.make_coord_scan(coords_fg)
@@ -103,30 +102,26 @@ class FilegroupScan():
     @property
     def variables(self) -> List[str]:
         """List of variables contained in this filegroup."""
-        csv = self.cs['var']
-        if csv.has_data():
-            v = csv[:].tolist()
-        else:
-            v = []
-        return v
+        cs = self.cs['var']
+        if cs.has_data():
+            return cs[:].tolist()
+        return []
 
     @property
     def contains(self) -> Dict[str, Optional[np.ndarray]]:
         """Index of values contained in this filegroup.
 
-        Indexed on available scope.
-        None designate a value not contained.
+        Indexed on available scope. None designate a value not contained.
         """
         out = {name: c.contains for name, c in self.cs.items()}
         return out
 
     def __repr__(self):
-        s = [self.__class__.__name__]
-        s.append(f"Name: {self.name}")
-        s.append(f"Root Directory: {self.root}")
-        s.append(f"Pre-regex: {self.pregex}")
-        s.append(f"Regex: {self.regex}")
-        s.append('')
+        s = [str(self),
+             f"Root directory: {self.root}",
+             f"Pre-regex: {self.pregex}",
+             f"Regex: {self.regex}",
+             ""]
 
         s.append("Coordinates for scan:")
         for name, cs in self.cs.items():
@@ -137,26 +132,25 @@ class FilegroupScan():
             s.append(''.join(s1))
         return '\n'.join(s)
 
-    def make_coord_scan(self, coords: List[Tuple[Coord, bool, str]]):
+    def __str__(self):
+        return "{}: {}".format(self.__class__.__name__, self.name)
+
+    def make_coord_scan(self, coords: CoordScanSpec):
         """Add CoordScan objects.
 
-        Each CoordScan is dynamically rebased
-        from its parent Coord.
-
-        :param coords: List of tuple containing the coordinate object,
-            the shared flag, and the name of the coordinate infile.
+        Each CoordScan is dynamically rebased from its parent Coord.
         """
         self.cs = {}
-        for coord, shared, name in coords:
-            cs = dlcs.get_coordscan(self, coord, shared, name)
-            self.cs.update({coord.name: cs})
+        for c in coords:
+            cs = get_coordscan(self, c.coord, c.shared, c.name)
+            self.cs.update({c.coord.name: cs})
 
     def iter_shared(self, shared: bool = None) -> Dict[str, CoordScan]:
         """Iter through CoordScan objects.
 
         :param shared: [opt] To iterate only shared coordinates (shared=True),
-            or only in coordinates (shared=False).
-            If left to None, iter all coordinates.
+            or only in coordinates (shared=False). If left to None, iter all
+            coordinates.
         """
         cs = {}
         for name, c in self.cs.items():
@@ -174,14 +168,18 @@ class FilegroupScan():
     def set_scan_regex(self, pregex: str, **replacements: str):
         """Specify the pre-regex.
 
-        Create a proper regex from the pre-regex.
-        Find the matchers: replace them by the appropriate regex,
-        store segments for easy replacement by the matches later.
+        Create a proper regex from the pre-regex. Find the matchers, replace
+        them by the appropriate regex, store segments for easy replacement by
+        the matches later.
 
         :param pregex: Pre-regex.
         :param replacements: Matchers to be replaced by a constant.
             The arguments names must match a matcher in the pre-regex.
 
+        :raises KeyError: A matcher has an unknown dimension.
+        :raises TypeError: A dimensions is not shared but has a matcher.
+        :raises IndexError: A dimension is shared but has not matcher.
+
         Example
         -------
         >>> pregex = "%(prefix)_%(time:value)"
@@ -194,17 +192,22 @@ class FilegroupScan():
 
         m = self.scan_pregex(pregex)
 
-        # Separations between segments
-        idx = 0
+        idx = -1  # If no matcher, we get n_matcher = 0, see below
         regex = pregex
         for idx, match in enumerate(m):
-            matcher = dlcs.Matcher(match, idx)
+            matcher = Matcher(match, idx)
+            if matcher.coord not in self.cs:
+                raise KeyError("{} has no scanning coordinate '{}'."
+                               .format(self.name, matcher.coord))
+            if not self.cs[matcher.coord].shared:
+                raise TypeError(f"'{matcher.coord}' is not shared but has a matcher.")
             self.cs[matcher.coord].add_matcher(matcher)
             regex = regex.replace(match.group(), '(' + matcher.rgx + ')')
 
         for name, cs in self.iter_shared(True).items():
             if len(cs.matchers) == 0:
-                raise RuntimeError(f"'{name}' has no matcher in the pre-regex.")
+                raise IndexError(f"'{name}' in filegroup '{self.name}' is "
+                                 "shared but has no matcher in the pre-regex.")
 
         self.n_matcher = idx + 1
         self.regex = regex
@@ -212,15 +215,12 @@ class FilegroupScan():
 
     @staticmethod
     def scan_pregex(pregex: str) -> Optional[Iterator[re.match]]:
-        """Scan pregex for matchers.
-
-        :param pregex: Pre-regex.
-        """
+        """Scan pregex for matchers. """
         regex = r"%\(([a-zA-Z]*):([a-zA-Z]*)(?P<cus>:custom=)?((?(cus)[^:]+:))(:?dummy)?\)"
         m = re.finditer(regex, pregex)
         return m
 
-    def find_segments(self, m: Optional[Iterator[re.match]]):
+    def find_segments(self, m: Iterator[re.match]):
         """Find segments in filename.
 
         Store result.
@@ -243,6 +243,9 @@ class FilegroupScan():
                   **kwargs: Any) -> File:
         """Open a file.
 
+        Exception are handled by Tomate (which should close the file in case
+        an exception is thrown).
+
         :param filename: File to open.
         :param mode: Mode for opening (read only, replace, append, ...)
         :param log_lvl: {'debug', 'info', 'warning'} Level to log the opening at.
@@ -255,39 +258,30 @@ class FilegroupScan():
 
     def is_to_open(self) -> bool:
         """Return if the current file has to be opened."""
-        to_open = (any([cs.is_to_open() for cs in self.cs.values()])
-                   or ('gen' in self.scan_attr and not self.scan_attr['gen'][1]))
+        to_open = (any(cs.is_to_open() for cs in self.cs.values())
+                   or any(s.kind == 'gen' for s in self.scanners))
         return to_open
 
     def scan_general_attributes(self, file: File):
         """Scan for general attributes."""
-        func, scanned, kwargs = self.scan_attr['gen']
-        if not scanned:
-            log.debug('Scanning file for general attributes.')
-            infos = func(self, file, **kwargs)
-            log.debug("Found infos %s", list(infos.keys()))
-            already_present = [info for info in infos if infos in self.vi.infos]
-            for info in already_present:
-                infos.pop(info)
-            self.vi.set_infos(**infos)
-
-            self.scan_attr['gen'][1] = True
+        for s in self.scanners:
+            if s.kind == 'gen' and s.to_scan:
+                log.debug('Scanning file for general attributes.')
+                infos = s.scan(self, file)
+                log.debug("Found infos %s", list(infos.keys()))
+                self.vi.set_infos_default(**infos)
 
     def scan_file(self, filename: str):
         """Scan a single file.
 
-        Match filename against regex.
-        If first match, retrieve segments.
+        Match filename against regex. If first match, retrieve segments.
 
-        If needed, open file.
-        Scan general attributes.
-        For all CoordScan, scan coordinate attributes,
-        scan values, and in-file indices.
+        If needed, open file. Scan general attributes. For all CoordScan, scan
+        coordinate attributes, scan elements.
 
         Close file.
         """
         m = re.match(self.regex, filename)
-
         filename = os.path.join(self.root, filename)
 
         # Discard completely non matching files
@@ -304,8 +298,7 @@ class FilegroupScan():
             file = self.open_file(filename, mode='r', log_lvl='debug')
 
         try:
-            if 'gen' in self.scan_attr and not self.scan_attr['gen'][1]:
-                self.scan_general_attributes(file)
+            self.scan_general_attributes(file)
 
             for cs in self.cs.values():
                 cs.scan_attributes(file)
@@ -322,20 +315,29 @@ class FilegroupScan():
     def find_files(self) -> List[str]:
         """Find files to scan.
 
-        Uses os.walk.
-        Sort files alphabetically
+        Uses os.walk. Limit search to `MAX_DEPTH_SCAN` levels of directories
+        deep.
 
-        :raises RuntimeError: If no files are found.
+        If `file_override` is set, bypass this search, just use it.
+
+        Sort files alphabetically.
+
+        :raises IndexError: If no files are found.
         """
-        # Using a generator should fast things up even though
-        # less readable
-        files = [os.path.relpath(os.path.join(root, file), self.root)
-                 for root, _, files in os.walk(self.root)
-                 for file in files]
+        if self.file_override:
+            return [self.file_override]
+
+        files = []
+        for root, _, files_ in os.walk(self.root):
+            depth = len(os.path.relpath(root, self.root).split(os.sep)) - 1
+            if depth > self.MAX_DEPTH_SCAN:
+                break
+            files += [os.path.relpath(os.path.join(root, file), self.root)
+                      for file in files_]
         files.sort()
 
         if len(files) == 0:
-            raise RuntimeError(f"No files were found in {self.root}")
+            raise IndexError(f"No files were found in {self.root}")
 
         log.debug("Found %s files in %s", len(files), self.root)
 
@@ -344,26 +346,21 @@ class FilegroupScan():
     def scan_files(self):
         """Scan files.
 
-        Reset scanning coordinate if they are to scan.
-        Find files.
-        Scan each file.
-        Set CoordScan values.
+        Reset scanning coordinate if they are to scan. Find files. Scan each
+        file. Set CoordScan elements.
 
         :raises NameError: If no files matching the regex were found.
         :raises ValueError: If no values were detected for a coordinate.
         """
-        for elt in self.scan_attr:
-            self.scan_attr[elt][1] = False
+        for s in self.scanners:
+            s.to_scan = True
         # Reset CoordScan
         for cs in self.cs.values():
-            if cs.is_to_scan():
-                cs.scanned = False
-                if cs.scan_attributes_func is not None:
-                    cs.scan_attr = True
-                if 'manual' not in cs.scan:
-                    cs.reset()
-                elif cs.shared:
-                    cs.matches = [[] for _ in range(len(cs.values))]
+            cs.scanned = False
+            for s in cs.scanners:
+                s.to_scan = True
+            if not cs.manual:
+                cs.reset()
         self.found_file = False
 
         files = self.find_files()
@@ -375,7 +372,11 @@ class FilegroupScan():
                             .format(self.name, self.regex))
 
         for cs in self.cs.values():
-            cs.set_values()
+            for elt, value in cs.fixed_elts.items():
+                cs.update_values(cs.values,
+                                 **{elt: [value for _ in range(len(cs.values))]})
+
+            cs.sort_elements()
 
             if (cs.coord.name != 'var'
                     and cs.units != '' and cs.coord.units != ''
@@ -389,35 +390,22 @@ class FilegroupScan():
                 try:
                     cs.values = f(cs.values, cs.units, cs.coord.units)
                 except NotImplementedError:
-                    log.warning("Units conversion should happen for '%s' (%s)"
-                                " from '%s' to '%s' but no function is defined.",
+                    log.warning("Units conversion should happen for '%s' (%s) "
+                                "from '%s' to '%s' but no function is defined.",
                                 cs.name, self.name, cs.units, cs.coord.units)
 
-            if cs.is_to_check() or cs.name == 'var':
-                if len(cs.values) == 0:
-                    raise ValueError("No values detected ({0}, {1})".format(
-                        cs.name, self.name))
-                cs.update_values(cs.values)
+            if len(cs.values) == 0:
+                raise ValueError("No values detected ({0}, {1})".format(
+                    cs.name, self.name))
+            cs.update_values(cs.values)
 
-    def set_scan_gen_attrs_func(self, func: Callable[..., Dict], **kwargs: Any):
-        """Set function for scanning general attributes.
-
-        :param func: Function that recovers variables attributes in file.
-            See scan_general_attributes_default() for a better
-            description of the function interface.
-        :param kwargs: [opt] Passed to the function.
-        """
-        self.scan_attr['gen'] = [func, False, kwargs]
-
-    def set_scan_var_attrs_func(self, func: Callable[..., Dict], **kwargs: Any):
-        """Set the function for scanning variables specific attributes.
-
-        :param func: Function that recovers variables attributes in file.
-            See scan_variables_attributes_default() for a better
-            description of the function interface.
-        :param kwargs: [opt] Passed to the function.
-        """
-        self.scan_attr['var'] = [func, False, kwargs]
+    def add_scan_attrs_func(self, func: Callable,
+                            kind: str = None, **kwargs: Any):
+        """Add a function for scanning attributes."""
+        func = Scanner(kind, func, **kwargs)
+        if func.kind not in ['var', 'gen']:
+            raise KeyError("Attributes scanner kind should be 'gen' or 'var'")
+        self.scanners.append(func)
 
     def apply_coord_selection(self):
         """Apply CoordScan selection."""
@@ -430,36 +418,8 @@ class FilegroupScan():
             cs.slice(key.no_int())
 
 
-def scan_general_attributes_default(fg: 'FilegroupLoad', file: File,
-                                    **kwargs: Any) -> Dict[str, Any]:
-    """Scan general attributes in file.
-
-    :param file: Object to access file.
-        The file is already opened by FilegroupSan.open_file().
-
-    :returns: Dictionnary of attributes.
-        {attribute name: attribute value}.
-        Attributes are added to the VI.
-    """
-    raise NotImplementedError()
-
-
-def scan_variables_attributes_default(fg: 'FilegroupLoad', file: File,
-                                      **kwargs: Any) -> Dict[str, Dict[str, Any]]:
-    """Scan variable specific attributes.
-
-    :param file: Object to access file.
-        The file is already opened by FilegroupScan.open_file().
-
-    :returns: Attributes per variable.
-        {variable name: {attribute name: value}}
-        Attributes are added to the VI.
-    """
-    raise NotImplementedError()
-
-
-def make_filegroup(fg_type: Type, root: str, dims: List[Coord],
-                   coords_fg: List[Tuple[Union[str, Coord], Union[str, bool], str]],
+def make_filegroup(fg_type: Type, root: str, dims: Dict[str, Coord],
+                   coords_fg: Iterable[CoordScanSpec],
                    vi: VariablesInfo,
                    root_fg: str = None, name: str = '',
                    variables_shared: bool = False,
@@ -467,46 +427,29 @@ def make_filegroup(fg_type: Type, root: str, dims: List[Coord],
     """Convenience function to create filegroup.
 
     :param fg_type: Class of filegroup to add. Dependant on the file-format.
-    :param root: Base root.
-    :param coords_fg: Coordinates used in this grouping of files.
-        Each element of the list is a tuple of length 2 or 3 with
-        the coordinate (or its name), a shared flag, and eventually
-        the name of the coordinate in the file.
-        The flag can be 'shared' or 'in', or a boolean (True = shared).
-        The name is optional, if not specified the name of the coordinate
-        object is used.
-        Variables dimension can be omitted.
+    :param root: Database root directory.
+    :param dims: List of parent coordinates.
+    :param coords_fg: Coordinates used in this grouping of files. Each
+        element is an :class:`CoordScanSpec
+        <tomate.filegroup.spec.CoordScanSpec>`. See its documentation
+        for details.
+    :param vi: Global VariablesInfo instance.
+    :param root_fg: Subfolder, relative to root.
     :param name: Name of the filegroup.
-    :param root_fg: [opt] Subfolder from root.
     :param variables_shared: [opt] If the Variables dimension is shared.
         Default is False.
-    :param kwargs: [opt] Passed to the fg_type initializator.
+    :param kwargs: [opt] Passed to the `fg_type` initializator.
     """
-    shared_corres = {'in': False, 'shared': True}
-    for i, c_fg in enumerate(coords_fg):
-        if isinstance(c_fg[0], str):
-            c_name = c_fg[0]
-            try:
-                c_fg[0] = dims[c_name]
-            except KeyError:
-                raise KeyError("'{}' is not in constructor dimensions.".format(c_name))
-        if len(c_fg) < 3:
-            c_fg.append(c_fg[0].name)
-        shared = c_fg[1]
-        if not isinstance(shared, bool):
-            if shared not in shared_corres:
-                raise ValueError("Shared must be bool or {}\n({}, {})"
-                                 .format(list(shared_corres.keys()),
-                                         name, c_fg[0].name))
-            shared = shared_corres[shared]
-        coords_fg[i][1] = shared
+    for css in coords_fg:
+        css.process(dims)
 
     if root_fg is None:
         root_fg = ''
     root_fg = os.path.join(root, root_fg)
 
-    if all([c[0].name != 'var' for c in coords_fg]):
-        coords_fg.insert(0, [dims['var'], variables_shared, 'var'])
+    coords_fg = list(coords_fg)
+    if all(css.coord.name != 'var' for css in coords_fg):
+        coords_fg.insert(0, CoordScanSpec(dims['var'], variables_shared, 'var'))
     fg = fg_type(root_fg, None, coords_fg, vi, name, **kwargs)
 
     return fg
